#!/usr/bin/ruby
# encoding: utf-8

class Record < Hash
  def castka
    self[:castka].to_i
  end
  def currencify
    castka.currencify
  end
end

module Session
  NEZNAMY_HERMAN='Neznámý'
  @@cena_herneho = {
    'studentské' => 40,
    'dětské' => 30,
    'dospělé' => 50,  
  }

  # abstraktni metody k dodefinovatni v potomcich
  def abstract; raise "Abstract method, please redefine in child object."; end
  def value(key,type_of_value=:castka); abstract; end   # vrati hodnotu z dat
  def values(key,type_of_value=nil); abstract; end  # jako value ale nevadi mu vic hodnot

  # pomocné class metody
  def self.sum_method(method_name,sum_attribute=:castka)
    define_method(method_name) do
      raise "Bad method name" unless /^(.*)_celkem$/.match(method_name.to_s)
      result = send($1)
      if result.is_a?(Array)
        result.collect{|item| item[sum_attribute].to_i}.sum
      elsif result.nil?
        0
      elsif result.is_a?(Record)
        result[sum_attribute].to_i
      else
        raise "Chyba pri scitani (#{result.inspect})"
      end
    end
  end
  def self.value(method_name, attribute=:castka, default=nil)
    define_method(method_name) do 
      result = value(method_name.to_sym,attribute)
      if attribute==:castka && !result.nil?
        begin
          result.val = result.val.to_i  # aby zachoval ref udaj ulozeny v proxy
        rescue 
          result = result.to_i
        end
      end
      result || default
    end
  end
  def self.record(method_name,attributes, options = {})
    define_method(method_name) do  ## neni to moc DRY, ten kod se da okopirovat z records
      rows = values(options[:typ_zaznamu] || method_name.to_sym,attributes+[:ref,:datum]).collect { |row|
        row[:castka] = row[:castka].to_i unless row[:castka].nil?
        Record.new.replace(row)
      }
      raise "Cekal jsem maximalne jeden zaznam typu #{method_name} ale je jich #{rows.size}" if rows.size>1
      rows.first
    end
  end
  def self.records(method_name,attributes, options = {})
    define_method(method_name) do
      values(options[:typ_zaznamu] || method_name.to_sym,attributes+[:ref,:datum]).collect { |row|
        row[:castka] = row[:castka].to_i unless row[:castka].nil?
        Record.new.replace(row)
      }
    end
  end

  # základni data
  value :datum, :datum
  value :balance_start
  value :balance_end
  records :obsluha, [:herman]
#  value :prodej_jidla_celkem,:castka, 0
#  record :prodej_jidla, [:castka], :typ_zaznamu => 'prodej_jidla_celkem'
  records :prodej_jidla, [:castka]
  records :konzumace_na_dluh,[:herman,:castka]
  records :konzumace_na_kredit,[:herman,:castka]
  records :herne,[:kategorie,:pocet, :castka, :castka_za_kus]
  records :volnasky,[:herman,:pocet]
  records :splatky,[:herman,:castka]
  records :ostatni,[:herman,:popis,:castka]
  records :prodej_her,[:herman,:popis,:castka,:hra,:kategorie,:paragon,:rezervace]
  records :prodej_her_bazar,[:popis,:castka,:hra,:paragon,]
    
  # souhrny
  sum_method :konzumace_na_kredit_celkem
  sum_method :konzumace_na_dluh_celkem
  sum_method :herne_celkem
  sum_method :ostatni_celkem
  sum_method :splatky_celkem
  sum_method :ostatni_prijmy_celkem
  sum_method :ostatni_vydaje_celkem
  sum_method :prodej_her_celkem
  sum_method :prodej_jidla_celkem
  sum_method :prodej_her_bazar_celkem

  # odvozená data
  def vydaj_jidla_celkem
    prodej_jidla_celkem+konzumace_na_dluh_celkem+konzumace_na_kredit_celkem
  end
  def prijem_za_jidlo_celkem
    prodej_jidla_celkem+splatky_celkem
  end
  def manko
    if balance_end.nil? || balance_end_theory.nil?
      nil
    else 
      balance_end-balance_end_theory
    end
  end
  def ostatni_prijmy
    ostatni.select{ |row| row[:castka]>=0 }
  end
  def ostatni_vydaje
    ostatni.select{ |row| row[:castka]<0 }.collect{|row| row[:castka]= -row[:castka];  row}
  end
  def prazdny_den?
    balance_start.empty_or_zero? &&  balance_end.empty_or_zero? &&  konzumace_na_kredit.empty? &&   herne.empty? &&
    prodej_jidla_celkem.empty_or_zero? &&  splatky.empty? && ostatni.empty? &&  prodej_her.empty? &&  obsluha.empty?
  end
  def balance_end_theory
    value = value(:balance_end_theory)
    if !balance_start.nil?
      sum = balance_start+prijem_hotovosti-vydaj_hotovosti
      raise "Celkove teoretické balance podle sheetu #{value} nesouhlasi se sumou #{sum} (=#{balance_start}+#{prijem_hotovosti}-#{vydaj_hotovosti} kde #{prijem_hotovosti}=#{herne_celkem}+#{prodej_jidla_celkem}+#{splatky_celkem}+#{ostatni_prijmy_celkem}+ #{prodej_her_celkem}+#{prodej_her_bazar_celkem}s)" if value!=sum
    end
    value
  end  
  def prijem_hotovosti
    (herne_celkem+prodej_jidla_celkem+splatky_celkem+ostatni_prijmy_celkem+ prodej_her_celkem+prodej_her_bazar_celkem).to_i
  end 
  def vydaj_hotovosti
    ostatni_vydaje_celkem
  end
  
end

# proxy class, ktery k hodnote uchovava jeji referenci zdroje
class Proxy
  instance_methods.each { |m| undef_method m unless m =~ /(^__|^send$|^object_id$)/ }
  attr_accessor :ref  
  attr_accessor :val
  
  def initialize(init={})
    self.ref = init[:ref]
    self.val = init[:val]
  end  
 protected
  def method_missing(name, *args, &block)
    val.send(name, *args, &block)
  end
end

class NilClass
  def self.strftime(a='')
    ''
  end
  def strftime(a='')
    ''
  end
end

class Array
 def uniq(&blk)
   blk ||= lambda {|x| x}
   already_seen = Array.new
   uniq_array = []

   self.each_with_index do |value, i|
     x = blk.call(value)
     unless already_seen.include? x
       already_seen << x
       uniq_array << value
     end
   end
   uniq_array
 end
 def compact
   self.select {|i| !i.nil?}
 end
end

class String
  def to_s
    self
  end
end
